"""The Django model instance factory."""

from __future__ import annotations

import warnings
from collections.abc import Callable, Iterable, Mapping
from typing import TYPE_CHECKING, Any, ClassVar, TypeVar, cast

from django.apps import apps as django_apps
from django.db import models

from .flags import *  # noqa: F403

if TYPE_CHECKING:
    from django.contrib.contenttypes.fields import GenericForeignKey
    from typing_extensions import TypeGuard

    from .typing import F, FieldSetter


def generate_default_value(field: F, unique: bool = False) -> FieldSetter[Any]:
    # The defaults module's imports use the Django settings so we
    # delay this import as long as possible
    from .defaults import generate_default_value as _generate_default_value

    return _generate_default_value(field, unique)


class ApplicationNamespaceDescriptor:

    """Descriptor for an application namespace on an instance factory.

    Each instance of this class dynamically generates a namespace
    class for the *models* its given.

    """

    def __init__(self, app_models: Iterable[type[models.Model]]) -> None:
        class ApplicationNamespace:
            def __init__(self, ifactory: InstanceFactory) -> None:
                self.ifactory = ifactory

        def make_create(model: type[models.Model]) -> Callable[..., models.Model]:
            def create(self: ApplicationNamespace, **attrs: Any) -> models.Model:
                return self.ifactory.create(model, **attrs)

            create.__name__ = model._meta.model_name  # type: ignore[assignment]
            return create

        for model in app_models:
            if model._meta.model_name:
                setattr(
                    ApplicationNamespace, model._meta.model_name, make_create(model)
                )

        self.namespace_class = ApplicationNamespace

    def __get__(
        self, obj: InstanceFactory, objtype: type[InstanceFactory] | None = None
    ) -> Any:
        return self.namespace_class(obj)


M = TypeVar("M", bound=models.Model)


class InstanceFactory:

    """A factory for Django model instances to use in unit tests."""

    Create = CreateRelatedInstance
    Lookup = LookupRelatedInstance
    Unique = Unique

    # Global defaults for all models in the format
    # {model_label_lower: {field_name: default_value, ...}, ...}.
    _defaults: ClassVar[dict[str, dict[str, Any]]] = {}

    @classmethod
    def register_all_models(cls) -> None:
        """Register all installed models for production at the factory."""
        if not django_apps.ready:
            raise RuntimeError(
                "InstanceFactory.register_models() must be called after Django has "
                "been initialized"
            )
        for app in django_apps.get_app_configs():
            models = list(app.get_models())
            for model in models:
                cls._set_initial_defaults(model)
            setattr(cls, app.label, ApplicationNamespaceDescriptor(models))

    @classmethod
    def _set_initial_defaults(cls, model: type[models.Model]) -> None:
        # Generate initial default values for the fields of a *model*.
        def needs_default(
            field: models.Field | models.ForeignObjectRel | GenericForeignKey,
        ) -> TypeGuard[models.Field]:
            return (
                isinstance(field, models.Field)
                and not field.many_to_many
                and not field.null
                and not field.has_default()
                and not isinstance(field, models.AutoField)
            )

        def generate_default(
            field: models.Field,
        ) -> CreateRelatedInstance | FieldSetter[Any]:
            if field.many_to_one or field.one_to_one:
                return InstanceFactory.Create
            return generate_default_value(field)

        cls._defaults[model._meta.label_lower] = {
            field.name: generate_default(field)
            for field in model._meta.get_fields()
            if needs_default(field)
        }

    @classmethod
    def configure_defaults(
        cls,
        model: str | type[models.Model],
        attrs: dict[str, Any] | None = None,
        /,
        **kwargs: Any,
    ) -> None:
        """Configure the default values for the given *model*.

        *model* can be an actual model class or a model label. *attrs*
        is deprecated. Use keyword arguments to map field names to
        default values.

        """
        if isinstance(attrs, Mapping) and not kwargs:
            warnings.warn(
                "Passing the default field values in a dictionary to "
                "ifactory.configure_defaults() is deprecated. Use kwargs instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            kwargs = attrs
        elif attrs is not None:
            kwargs["attrs"] = attrs

        if isinstance(model, str):
            model = django_apps.get_model(model)
        if not cls._defaults:  # make sure the factory is initialized
            cls.register_all_models()

        def validate_attr(name: str, value: Any) -> Any:
            field = model._meta.get_field(name)
            if not isinstance(field, models.Field):
                raise TypeError(
                    f"Can't configure default value for `{model._meta.label}.{name}` "
                    f"since it's either a reverse or a generic relationship"
                )
            if (is_create(value) or is_lookup(value)) and field.related_model is None:
                raise TypeError(
                    f"Can't create related instance for `{model._meta.label}.{name}` "
                    f"since it's not a field with a related model"
                )
            if is_unique(value):
                value = generate_default_value(field, unique=True)
            return (name, value)

        cls._defaults[model._meta.label_lower].update(
            validate_attr(name, value) for name, value in kwargs.items()
        )

    def __init__(self) -> None:
        if not self._defaults:  # make sure the factory is initialized
            self.register_all_models()

    def create(
        self, model: type[M], attrs: dict[str, Any] | None = None, /, **kwargs: Any
    ) -> M:
        """Create an instance of the given model."""
        if isinstance(attrs, Mapping) and not kwargs:
            warnings.warn(
                "Passing the model instance field values in a dictionary to "
                "ifactory.create() is deprecated. Use kwargs instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            kwargs = attrs
        elif attrs is not None:
            kwargs["attrs"] = attrs

        def get_default_value(name: str, value: Any) -> Any:
            # _set_initial_defaults() and configure_defaults() should
            # make sure that there's only valid attributes
            # (corresponding to models.Field) in self._defaults.  They
            # should also make sure that Create and Lookup defaults
            # have a related model.
            field = cast(models.Field, model._meta.get_field(name))
            if value is self.Create:
                related_model = cast("type[models.Model]", field.related_model)
                value = self.create(related_model)
            elif isinstance(value, self.Create):
                related_model = cast("type[models.Model]", field.related_model)
                value = self.create(related_model, **value.attrs)
            elif value is self.Lookup:
                related_model = cast("type[models.Model]", field.related_model)
                value = related_model.objects.first()
            elif isinstance(value, self.Lookup):
                related_model = cast("type[models.Model]", field.related_model)
                value = related_model.objects.get(**value.attrs)
            elif callable(value):
                value = value()
            return value

        defaults = {
            k: get_default_value(k, v)
            for k, v in self._defaults[model._meta.label_lower].items()
            if k not in kwargs
        }
        defaults.update(kwargs)
        instance = model(**defaults)
        instance.clean()
        instance.save()
        return instance



"""Pytest plugin module."""

from __future__ import annotations

import pytest

from .ifactory import InstanceFactory


class PytestDjangoIFactorySpec:

    """Hook specification namespace for this plugin."""

    def pytest_django_ifactory_configure(self, ifactory: type[InstanceFactory]) -> None:
        """Configure model field defaults at the pytest instance factory."""


def pytest_addhooks(pluginmanager: pytest.PytestPluginManager) -> None:
    pluginmanager.add_hookspecs(PytestDjangoIFactorySpec)


@pytest.fixture(autouse=True, scope="session")
def _django_ifactory_register(request: pytest.FixtureRequest) -> None:
    request.config.pluginmanager.hook.pytest_django_ifactory_configure(
        ifactory=InstanceFactory
    )


@pytest.fixture
def ifactory(db: None) -> InstanceFactory:  # noqa: ARG001
    return InstanceFactory()


@pytest.fixture
def transactional_ifactory(transactional_db: None) -> InstanceFactory:  # noqa: ARG001
    return InstanceFactory()



"""Type helpers for pytest-django-ifactory."""

from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING, Any, TypeVar, Union

if TYPE_CHECKING:
    from django.db.models import Field

T = TypeVar("T")
FieldSetter = Union[T, Callable[[], T], None]

F = TypeVar("F", bound="Field[Any, Any]")
ValueGenerator = Callable[[F, bool], FieldSetter[Any]]



"""Unit and regression tests for the instance factory class."""

from __future__ import annotations

import copy
from typing import TYPE_CHECKING

import pytest

from testapp.models import AllFieldsModel, ModelA, ModelB, ModelC

if TYPE_CHECKING:
    from collections.abc import Iterator

    from pytest_django_ifactory.ifactory import InstanceFactory


@pytest.fixture(params=["ifactory", "transactional_ifactory"])
def ifactory(request: pytest.FixtureRequest) -> InstanceFactory:
    return request.getfixturevalue(request.param)  # type: ignore[no-any-return]


def test_all_fields(ifactory: InstanceFactory) -> None:
    instance = ifactory.create(AllFieldsModel)
    for field in instance._meta.get_fields():
        assert getattr(instance, field.name) is not None


def test_initialization(ifactory: InstanceFactory) -> None:
    assert hasattr(ifactory, "testapp")
    assert hasattr(ifactory.testapp, "modela")
    assert hasattr(ifactory.testapp, "modelb")
    assert isinstance(ifactory.testapp.modela().name, str)
    b = ifactory.testapp.modelb()
    assert isinstance(b.name, str)
    assert isinstance(b.required_a, ModelA)
    assert b.nullable_a1 is None
    assert b.nullable_a2 is None


def test_blank_default(ifactory: InstanceFactory) -> None:
    a = ifactory.create(ModelA)
    assert not a.blank


def test_create_with_attrs(ifactory: InstanceFactory) -> None:
    assert ifactory.create(ModelA, name="foo").name == "foo"


def test_create_with_model_attr(ifactory: InstanceFactory) -> None:
    c = ifactory.create(ModelC, model="a field named model")
    assert c.model == "a field named model"


class TestConfigure:
    @pytest.fixture
    def cifactory(self, ifactory: InstanceFactory) -> Iterator[InstanceFactory]:
        original_defaults = copy.deepcopy(ifactory._defaults)
        yield ifactory
        type(ifactory)._defaults = original_defaults

    @pytest.mark.parametrize("model", ["testapp.modelb", ModelB])
    def test_primitive(
        self, cifactory: InstanceFactory, model: str | type[ModelB]
    ) -> None:
        cifactory.configure_defaults(model, name="foo")
        b = cifactory.create(ModelB)
        assert b.name == "foo"

    def test_callable(self, cifactory: InstanceFactory) -> None:
        cifactory.configure_defaults("testapp.modela", name=lambda: "foo")
        assert cifactory.create(ModelA).name == "foo"

    @pytest.mark.parametrize("model", ["testapp.NoModel", "noapp.NoModel"])
    def test_invalid_model(self, cifactory: InstanceFactory, model: str) -> None:
        with pytest.raises(LookupError):
            cifactory.configure_defaults(model)

    def test_create(self, cifactory: InstanceFactory) -> None:
        cifactory.configure_defaults("testapp.modelb", nullable_a1=cifactory.Create)
        assert isinstance(cifactory.create(ModelB).nullable_a1, ModelA)

    def test_create_with_attrs(self, cifactory: InstanceFactory) -> None:
        cifactory.configure_defaults(
            "testapp.modelb", nullable_a1=cifactory.Create(name="foo")
        )
        instance = cifactory.create(ModelB)
        assert instance.nullable_a1
        assert instance.nullable_a1.name == "foo"

    def test_lookup(self, cifactory: InstanceFactory) -> None:
        ModelA.objects.create(name="foo")
        ModelA.objects.create(name="bar")
        cifactory.configure_defaults("testapp.modelb", nullable_a1=cifactory.Lookup)
        instance = cifactory.create(ModelB)
        assert instance.nullable_a1
        assert instance.nullable_a1.name == "foo"

    def test_lookup_with_attrs(self, cifactory: InstanceFactory) -> None:
        ModelA.objects.create(name="foo")
        cifactory.configure_defaults(
            "testapp.modelb", nullable_a1=cifactory.Lookup(name="foo")
        )
        instance = cifactory.create(ModelB)
        assert instance.nullable_a1
        assert instance.nullable_a1.name == "foo"

    def test_unique(self, cifactory: InstanceFactory) -> None:
        cifactory.configure_defaults("testapp.modela", category=cifactory.Unique)
        instance1 = cifactory.create(ModelA)
        instance2 = cifactory.create(ModelA)
        assert instance1.category != instance2.category

    def test_configure_model_attr(self, cifactory: InstanceFactory) -> None:
        cifactory.configure_defaults("testapp.modelc", model="model default")
        instance = cifactory.create(ModelC)
        assert instance.model == "model default"

    def test_plugin(self, cifactory: InstanceFactory) -> None:
        cifactory.configure_defaults("testapp.ModelA", name="adam")
        cifactory.configure_defaults(
            "testapp.ModelB", name="bert", nullable_a1=cifactory.Create
        )

        a = cifactory.create(ModelA, name="alan")
        assert ModelA.objects.count() == 1
        assert a.name == "alan"
        b = cifactory.create(ModelB, required_a=a)
        assert ModelB.objects.count() == 1
        assert ModelA.objects.count() == 2
        assert b.name == "bert"
        assert b.required_a is a
        assert isinstance(b.nullable_a1, ModelA)
        assert b.nullable_a1.name == "adam"
        assert b.nullable_a2 is None



"""Test configuration for pytest-django-ifactory."""

from __future__ import annotations

import os
import shutil
from pathlib import Path

import pytest
from django.conf import settings

pytest_plugins = ["pytester"]


def pytest_configure() -> None:
    settings.configure(
        DATABASES={"default": {"ENGINE": "django.contrib.gis.db.backends.spatialite"}},
        SPATIALITE_LIBRARY_PATH=os.environ.get("SPATIALITE_LIBRARY_PATH"),
        INSTALLED_APPS=[
            "django.contrib.auth",
            "django.contrib.contenttypes",
            "django.contrib.gis",
            "testapp",
        ],
        USE_TZ=True,
    )


@pytest.fixture
def django_pytester(pytester: pytest.Pytester) -> pytest.Pytester:
    cwd = Path(__file__).parent
    shutil.copy(cwd / "djangosettings.py", pytester.path)
    shutil.copytree(cwd / "testapp", pytester.path / "testapp")
    return pytester



"""Untyped unit and regression tests for the instance factory class.

This tests the older API that doesn't work well with static type
checking.

"""

from testapp.models import ModelA, ModelB


def test_create(ifactory):
    instance_a = ifactory.testapp.modela(name="model a")
    assert isinstance(instance_a, ModelA)
    assert instance_a.name == "model a"
    instance_b = ifactory.testapp.modelb(name="model b")
    assert isinstance(instance_b, ModelB)
    assert instance_b.name == "model b"


def test_dynamic_method_names(ifactory):
    assert ifactory.testapp.modela.__name__ == "modela"
    assert ifactory.testapp.modelb.__name__ == "modelb"



"""Creation flags for the factory."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from typing_extensions import TypeGuard


__all__ = [
    "CreateRelatedInstance",
    "LookupRelatedInstance",
    "Unique",
    "is_create",
    "is_lookup",
    "is_unique",
]


class CreateRelatedInstance:

    """Flag for creating a related instance using the factory.

    Any *attrs* given to the constructor will be used when creating
    the related instance.

    """

    def __init__(self, **attrs: Any) -> None:
        self.attrs = attrs


def is_create(
    obj: Any,
) -> TypeGuard[CreateRelatedInstance | type[CreateRelatedInstance]]:
    return is_class_or_instance(obj, CreateRelatedInstance)


class LookupRelatedInstance:

    """Flag for looking up a related instance in the database.

    The *attrs* given to the constructor must uniquely identify the
    related instance in the database.

    """

    def __init__(self, **attrs: Any) -> None:
        self.attrs = attrs


def is_lookup(
    obj: Any,
) -> TypeGuard[LookupRelatedInstance | type[LookupRelatedInstance]]:
    return is_class_or_instance(obj, LookupRelatedInstance)


class Unique:

    """Flag for generating a unique value for a field."""


def is_unique(obj: Any) -> TypeGuard[Unique | type[Unique]]:
    return is_class_or_instance(obj, Unique)


def is_class_or_instance(obj: Any, cls: type[Any]) -> bool:
    return obj is cls or isinstance(obj, cls)



"""Default value generation for model fields."""

from __future__ import annotations

import datetime
import itertools
import string
from typing import TYPE_CHECKING, Any, overload

from django.core.exceptions import ImproperlyConfigured

if TYPE_CHECKING:
    from collections.abc import Callable

    from .typing import F, FieldSetter, ValueGenerator

try:
    from django.contrib.gis.db import models

    HAVE_GIS_MODELS = True
except ImproperlyConfigured:
    from django.db import models  # type: ignore[no-redef]

    HAVE_GIS_MODELS = False

_generators: dict[type[models.Field], ValueGenerator[Any]] = {}


@overload
def register(field_class: type[F], generator: ValueGenerator[F]) -> None:
    ...


@overload
def register(
    field_class: type[F], generator: None = ...
) -> Callable[[ValueGenerator[F]], None]:
    ...


def register(
    field_class: type[F], generator: ValueGenerator[F] | None = None
) -> Callable[[ValueGenerator[F]], None] | None:
    """Register a default value generator for a field class."""

    def decorator(generator: ValueGenerator[F]) -> None:
        _generators[field_class] = generator

    if generator is None:
        return decorator

    decorator(generator)
    return None


def generate_default_value(field: F, unique: bool = False) -> FieldSetter[Any]:
    """Generate a default value for a model field."""

    def _default_generator(_: models.Field[None, Any], unique: bool = False) -> None:
        if unique:
            raise ValueError("unique not supported")
        return None  # noqa: PLR1711, RET501

    generator = _generators.get(type(field), _default_generator)
    return generator(field, unique)


@register(models.BooleanField)
def boolean_field_default(field: models.BooleanField, unique: bool = False) -> bool:
    if unique or field.unique:
        raise ValueError("unique not supported")
    return True


def char_field_default(
    field: models.CharField | models.FileField | models.TextField, unique: bool = False
) -> FieldSetter[str]:
    length = min(4, field.max_length or 4)
    if unique or field.unique:
        it = itertools.permutations(string.ascii_letters, length)
        return lambda: "".join(next(it))
    return "" if field.blank else "abcd"[:length]


register(models.CharField, char_field_default)
register(models.EmailField, char_field_default)
register(models.FileField, char_field_default)
register(models.ImageField, char_field_default)
register(models.SlugField, char_field_default)
register(models.TextField, char_field_default)


@register(models.DateField)
def date_field_default(
    field: models.DateField, unique: bool = False
) -> FieldSetter[datetime.date]:
    if field.auto_now or field.auto_now_add:
        return None
    if unique or field.unique:
        it = itertools.count(step=24 * 60**2)
        return lambda: datetime.date.fromtimestamp(next(it))
    return datetime.date.fromtimestamp(0)


@register(models.DateTimeField)
def date_time_field_default(
    field: models.DateTimeField, unique: bool = False
) -> FieldSetter[datetime.datetime]:
    def todatetime(timestamp: int) -> datetime.datetime:
        return datetime.datetime.utcfromtimestamp(timestamp).replace(
            tzinfo=datetime.timezone.utc
        )

    if field.auto_now or field.auto_now_add:
        return None
    if unique or field.unique:
        it = itertools.count(step=60**2)
        return lambda: todatetime(next(it))
    return todatetime(0)


@register(models.FloatField)
def float_field_default(
    field: models.FloatField, unique: bool = False
) -> FieldSetter[float]:
    if unique or field.unique:
        it = itertools.count()
        return lambda: float(next(it))
    return 0.0


def integer_field_default(
    field: models.IntegerField | models.DecimalField, unique: bool = False
) -> FieldSetter[int]:
    if unique or field.unique:
        it = itertools.count()
        return lambda: next(it)
    return 0


register(models.BigIntegerField, integer_field_default)
register(models.DecimalField, integer_field_default)
register(models.IntegerField, integer_field_default)
register(models.PositiveIntegerField, integer_field_default)
register(models.PositiveSmallIntegerField, integer_field_default)
register(models.SmallIntegerField, integer_field_default)


@register(models.JSONField)
def json_field_default(field: models.JSONField, unique: bool = False) -> dict[Any, Any]:
    if unique or field.unique:
        raise ValueError("unique not supported")
    return {}


@register(models.TimeField)
def time_field_default(
    field: models.TimeField, unique: bool = False
) -> FieldSetter[datetime.time]:
    def totime(timestamp: int) -> datetime.time:
        return datetime.datetime.utcfromtimestamp(timestamp).time()

    if field.auto_now or field.auto_now_add:
        return None
    if unique or field.unique:
        it = itertools.count()
        return lambda: totime(next(it))
    return totime(0)


def make_geometry_field_default(template: str) -> ValueGenerator[models.GeometryField]:
    def _default(field: models.GeometryField, unique: bool = False) -> FieldSetter[str]:
        if unique or field.unique:
            it = itertools.count(1)
            return lambda: template % next(it)
        return template % 1

    return _default


if HAVE_GIS_MODELS:
    GIS_FIELDS: list[tuple[type[models.GeometryField], str]] = [
        (models.GeometryField, "POINT (%d 0)"),
        (models.PointField, "POINT (%d 0)"),
        (models.LineStringField, "LINESTRING (%d 0, 0 1)"),
        (models.PolygonField, "POLYGON ((0 0, %d 1, 0 1, 0 0))"),
        (models.MultiPointField, "MULTIPOINT ((%d 0))"),
        (models.MultiLineStringField, "MULTILINESTRING ((%d 0, 0 1))"),
        (models.MultiPolygonField, "MULTIPOLYGON (((0 0, %d 1, 0 1, 0 0)))"),
        (models.GeometryCollectionField, "GEOMETRYCOLLECTION (POINT (%d 0))"),
    ]

    for fieldtype, template in GIS_FIELDS:
        register(fieldtype, make_geometry_field_default(template))



"""Unit tests for pytest-django-ifactory's default value generators."""

from __future__ import annotations

import datetime

import pytest
from django.contrib.gis.db import models
from django.utils.dateparse import parse_datetime, parse_time

from pytest_django_ifactory.defaults import generate_default_value


def test_default_default() -> None:
    assert generate_default_value(models.Field()) is None


def test_default_unique_default() -> None:
    with pytest.raises(ValueError, match="unique not supported"):
        generate_default_value(models.Field(), unique=True)


@pytest.mark.parametrize("field_type", [models.BooleanField, models.JSONField])
@pytest.mark.parametrize("field_is_unique", [True, False])
def test_unique_not_implemented(
    field_type: type[models.Field], field_is_unique: bool
) -> None:
    field = field_type(unique=field_is_unique)
    with pytest.raises(ValueError, match="unique not supported"):
        generate_default_value(field, unique=not field_is_unique)


def test_boolean_field_default() -> None:
    assert generate_default_value(models.BooleanField()) is True


@pytest.mark.parametrize(
    "field_type",
    [
        models.CharField,
        models.EmailField,
        models.FileField,
        models.ImageField,
        models.SlugField,
        models.TextField,
    ],
)
class TestCharFieldDefault:
    def test_nonunique(self, field_type: type[models.Field]) -> None:
        field = field_type(max_length=9, unique=False)
        assert generate_default_value(field) == "abcd"

    @pytest.mark.parametrize("field_is_unique", [True, False])
    def test_unique(
        self, field_type: type[models.Field], field_is_unique: bool
    ) -> None:
        field = field_type(max_length=9, unique=field_is_unique)
        default = generate_default_value(field, unique=not field_is_unique)
        assert callable(default)
        assert default() == "abcd"
        assert default() == "abce"

    def test_short_max_length(self, field_type: type[models.Field]) -> None:
        field = field_type(max_length=3, unique=False)
        assert generate_default_value(field) == "abc"


class TestDateFieldDefault:
    def test_nonunique(self) -> None:
        default = generate_default_value(models.DateField())
        assert default == datetime.date.fromtimestamp(0)

    @pytest.mark.parametrize("field_is_unique", [True, False])
    def test_unique(self, field_is_unique: bool) -> None:
        field = models.DateField(unique=field_is_unique)  # type: ignore[var-annotated]
        default = generate_default_value(field, unique=not field_is_unique)
        assert callable(default)
        assert default() == datetime.date.fromtimestamp(0)
        assert default() == datetime.date.fromtimestamp(0) + datetime.timedelta(days=1)

    def test_auto_now(self) -> None:
        assert generate_default_value(models.DateField(auto_now=True)) is None

    def test_auto_now_add(self) -> None:
        assert generate_default_value(models.DateField(auto_now_add=True)) is None


class TestDateTimeFieldDefault:
    def test_nonunique(self) -> None:
        default = generate_default_value(models.DateTimeField())
        assert default == parse_datetime("1970-01-01T00:00Z")

    @pytest.mark.parametrize("field_is_unique", [True, False])
    def test_unique(self, field_is_unique: bool) -> None:
        field = models.DateTimeField(  # type: ignore[var-annotated]
            unique=field_is_unique
        )
        default = generate_default_value(field, unique=not field_is_unique)
        assert callable(default)
        assert default() == parse_datetime("1970-01-01T00:00Z")
        assert default() == parse_datetime("1970-01-01T01:00Z")

    def test_auto_now(self) -> None:
        assert generate_default_value(models.DateTimeField(auto_now=True)) is None

    def test_auto_now_add(self) -> None:
        assert generate_default_value(models.DateTimeField(auto_now_add=True)) is None


class TestFloatFieldDefault:
    def test_nonunique(self) -> None:
        assert generate_default_value(models.FloatField()) == pytest.approx(0)

    @pytest.mark.parametrize("field_is_unique", [True, False])
    def test_unique(self, field_is_unique: bool) -> None:
        field = models.FloatField(unique=field_is_unique)  # type: ignore[var-annotated]
        default = generate_default_value(field, unique=not field_is_unique)
        assert callable(default)
        assert default() == pytest.approx(0)
        assert default() == pytest.approx(1)


@pytest.mark.parametrize(
    "field_type",
    [
        models.BigIntegerField,
        models.DecimalField,
        models.IntegerField,
        models.PositiveIntegerField,
        models.PositiveSmallIntegerField,
        models.SmallIntegerField,
    ],
)
class TestIntegerFieldDefault:
    def test_nonunique(self, field_type: type[models.Field]) -> None:
        field = field_type()
        assert generate_default_value(field) == 0

    @pytest.mark.parametrize("field_is_unique", [True, False])
    def test_unique(
        self, field_type: type[models.Field], field_is_unique: bool
    ) -> None:
        field = field_type(unique=field_is_unique)
        default = generate_default_value(field, unique=not field_is_unique)
        assert callable(default)
        assert default() == 0
        assert default() == 1


class TestTimeFieldDefault:
    def test_nonunique(self) -> None:
        default = generate_default_value(models.TimeField())
        assert default == parse_time("00:00")

    @pytest.mark.parametrize("field_is_unique", [True, False])
    def test_unique(self, field_is_unique: bool) -> None:
        field = models.TimeField(unique=field_is_unique)  # type: ignore[var-annotated]
        default = generate_default_value(field, unique=not field_is_unique)
        assert callable(default)
        assert default() == parse_time("00:00:00")
        assert default() == parse_time("00:00:01")

    def test_auto_now(self) -> None:
        assert generate_default_value(models.TimeField(auto_now=True)) is None

    def test_auto_now_add(self) -> None:
        assert generate_default_value(models.TimeField(auto_now_add=True)) is None


@pytest.mark.parametrize(
    "field_type, expected_defaults",
    [
        (models.GeometryField, ["POINT (1 0)", "POINT (2 0)"]),
        (models.PointField, ["POINT (1 0)", "POINT (2 0)"]),
        (models.LineStringField, ["LINESTRING (1 0, 0 1)", "LINESTRING (2 0, 0 1)"]),
        (
            models.PolygonField,
            ["POLYGON ((0 0, 1 1, 0 1, 0 0))", "POLYGON ((0 0, 2 1, 0 1, 0 0))"],
        ),
        (models.MultiPointField, ["MULTIPOINT ((1 0))", "MULTIPOINT ((2 0))"]),
        (
            models.MultiLineStringField,
            ["MULTILINESTRING ((1 0, 0 1))", "MULTILINESTRING ((2 0, 0 1))"],
        ),
        (
            models.MultiPolygonField,
            [
                "MULTIPOLYGON (((0 0, 1 1, 0 1, 0 0)))",
                "MULTIPOLYGON (((0 0, 2 1, 0 1, 0 0)))",
            ],
        ),
        (
            models.GeometryCollectionField,
            ["GEOMETRYCOLLECTION (POINT (1 0))", "GEOMETRYCOLLECTION (POINT (2 0))"],
        ),
    ],
)
class TestGeometryFieldDefault:
    def test_nonunique(
        self, field_type: type[models.GeometryField], expected_defaults: list[str]
    ) -> None:
        default = generate_default_value(field_type())
        assert default == expected_defaults[0]

    @pytest.mark.parametrize("field_is_unique", [True, False])
    def test_unique(
        self,
        field_type: type[models.GeometryField],
        expected_defaults: list[str],
        field_is_unique: bool,
    ) -> None:
        field = field_type(unique=field_is_unique)
        default = generate_default_value(field, unique=not field_is_unique)
        assert callable(default)
        assert default() == expected_defaults[0]
        assert default() == expected_defaults[1]


def test_json_field_default() -> None:
    field = models.JSONField()
    assert generate_default_value(field) == {}



"""Unit and regression tests for the pytest-django-ifactory plugin."""

import pytest


def test_register_and_create(django_pytester: pytest.Pytester) -> None:
    django_pytester.makeconftest(
        """
        def pytest_django_ifactory_configure(ifactory):
            ifactory.configure_defaults('testapp.modela', {
                'name': 'adam',
            })
            ifactory.configure_defaults('testapp.modelb', {
                'name': 'bert',
                'nullable_a1': ifactory.Create,
            })
    """
    )
    django_pytester.makepyfile(
        """
        import pytest
        from testapp.models import ModelA, ModelB

        @pytest.fixture(params=["ifactory", "transactional_ifactory"])
        def ifactory(request):
            return request.getfixturevalue(request.param)

        def test_create(ifactory):
            a = ifactory.testapp.modela(name='alan')
            assert ModelA.objects.count() == 1
            assert a.name == 'alan'
            b = ifactory.testapp.modelb(required_a=a)
            assert ModelA.objects.count() == 2
            assert ModelB.objects.count() == 1
            assert b.name == 'bert'
            assert b.required_a is a
            assert isinstance(b.nullable_a1, ModelA)
            assert b.nullable_a1.name == 'adam'
            assert b.nullable_a2 is None
    """
    )
    result = django_pytester.runpytest_subprocess("--ds=djangosettings", "-v")
    result.assert_outcomes(passed=2)



"""Test cases for the code examples in the README."""

from __future__ import annotations

from itertools import groupby
from operator import methodcaller
from typing import TYPE_CHECKING

from django.contrib.auth.models import User

if TYPE_CHECKING:
    import pytest

    from pytest_django_ifactory import InstanceFactory


def get_duplicate_names() -> list[str]:
    all_users = User.objects.order_by("last_name", "first_name")
    users_by_name = groupby(all_users, methodcaller("get_full_name"))
    return [full_name for full_name, users in users_by_name if len(list(users)) > 1]


def test_get_duplicate_names(ifactory: InstanceFactory) -> None:
    ifactory.create(User, first_name="Albert", last_name="Einstein")
    ifactory.create(User, first_name="Albert", last_name="Einstein")
    ifactory.create(User, first_name="Erwin", last_name="Schrodinger")
    ifactory.create(User, first_name="Richard", last_name="Feynman")
    assert get_duplicate_names() == ["Albert Einstein"]
    assert User.objects.count() == 4


def test_register_and_create(django_pytester: pytest.Pytester) -> None:
    django_pytester.makeconftest(
        """
        def pytest_django_ifactory_configure(ifactory):
            ifactory.configure_defaults("auth.user", {
                "first_name": "Albert",
                "last_name": "Einstein",
            })
        """
    )
    django_pytester.makepyfile(
        """
        def test_albert_by_default(ifactory):
            albert = ifactory.auth.user()
            assert albert.get_full_name() == "Albert Einstein"
            not_albert = ifactory.auth.user(first_name="Erwin", last_name="Schrodinger")
            assert not_albert.get_full_name() == "Erwin Schrodinger"
       """
    )
    result = django_pytester.runpytest_subprocess("--ds=djangosettings", "-v")
    result.assert_outcomes(passed=1)



"""Django settings for pytest-django-ifactory's unit tests."""

import os

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = "0ay!$_8q2x7fo1bza(@1ovtudbu&1x=!7tzqa#a$cc8x@4o0m3"  # noqa: S105

INSTALLED_APPS = [
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.gis",
    "testapp",
]

DATABASES = {
    "default": {
        "ENGINE": "django.contrib.gis.db.backends.spatialite",
        "NAME": "pytestdjangoifactory.sqlite3",
    }
}

SPATIALITE_LIBRARY_PATH = os.environ.get("SPATIALITE_LIBRARY_PATH")



# Generated by Django 4.2.4 on 2023-08-27 13:23

import django.contrib.gis.db.models.fields
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="AllFieldsModel",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("boolean", models.BooleanField()),
                ("char", models.CharField(max_length=32)),
                ("email", models.EmailField(max_length=254)),
                ("file", models.FileField(upload_to="")),
                ("image", models.ImageField(upload_to="")),
                ("slug", models.SlugField()),
                ("text", models.TextField()),
                ("date", models.DateField()),
                ("datetime", models.DateTimeField()),
                ("float", models.FloatField()),
                ("biginteger", models.BigIntegerField()),
                ("decimal", models.DecimalField(decimal_places=0, max_digits=9)),
                ("integer", models.IntegerField()),
                ("positiveinteger", models.PositiveIntegerField()),
                ("positivesmallinteger", models.PositiveSmallIntegerField()),
                ("smallinteger", models.SmallIntegerField()),
                ("time", models.TimeField()),
                (
                    "geometry",
                    django.contrib.gis.db.models.fields.GeometryField(srid=4326),
                ),
                ("point", django.contrib.gis.db.models.fields.PointField(srid=4326)),
                (
                    "linestring",
                    django.contrib.gis.db.models.fields.LineStringField(srid=4326),
                ),
                (
                    "polygon",
                    django.contrib.gis.db.models.fields.PolygonField(srid=4326),
                ),
                (
                    "multipoint",
                    django.contrib.gis.db.models.fields.MultiPointField(srid=4326),
                ),
                (
                    "multilinestring",
                    django.contrib.gis.db.models.fields.MultiLineStringField(srid=4326),
                ),
                (
                    "multipolygon",
                    django.contrib.gis.db.models.fields.MultiPolygonField(srid=4326),
                ),
                (
                    "geometrycollection",
                    django.contrib.gis.db.models.fields.GeometryCollectionField(
                        srid=4326
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="ModelA",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=64, unique=True)),
                ("category", models.CharField(max_length=64)),
                ("blank", models.CharField(blank=True, max_length=64)),
            ],
        ),
        migrations.CreateModel(
            name="ModelC",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("model", models.CharField(max_length=64)),
            ],
        ),
        migrations.CreateModel(
            name="ModelB",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=64, unique=True)),
                (
                    "nullable_a1",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="testapp.modela",
                    ),
                ),
                (
                    "nullable_a2",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="testapp.modela",
                    ),
                ),
                (
                    "required_a",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="testapp.modela"
                    ),
                ),
            ],
        ),
    ]
